"""Pie plot functions."""

from __future__ import annotations

import matplotlib as mpl

from puma.plot_base import PlotBase
from puma.utils import get_good_pie_colours, logger


class PiePlot(PlotBase):
    """Pie plot class."""

    def __init__(
        self,
        wedge_sizes,
        colours: list | None = None,
        colour_scheme: str | None = None,
        labels: list | None = None,
        draw_legend: bool = False,
        mpl_pie_kwargs: dict | None = None,
        **kwargs,
    ):
        """Initialise the pie plot.

        Parameters
        ----------
        wedge_sizes : 1D array like
            The size of the wedges. Will be translated into the fractions automatically.
            So they don't have to add up to 1 or 100. The fractional area of each
            wedge is given by x/sum(x).
        colours : list, optional
            List of colours for the separate wedges. You have to specify as many
            colours as you have wedges. Instead, you can also specify a colour scheme
            with the `colour_scheme` argument, by default None
        colour_scheme : str, optional
            Name of the colour schemes as defined in puma.utils.get_good_pie_colours,
            by default None
        labels : list, optional
            A sequence of strings providing the labels for each wedge, by default None
        draw_legend : bool, optional
            If True, a legend will be drawn on the right side of the plot.
            If False, the labels will be drawn directly to the wedges. By default True
        mpl_pie_kwargs : dict, optional
            Keyword arguments that are handed to the matplotlib.pyplot.pie function.
            All arguments are allowed, except [`x`, `labels`, `colors`], by default None
        **kwargs : kwargs
            Keyword arguments from `puma.PlotObject`
        """
        super().__init__(vertical_split=draw_legend, **kwargs)
        self.wedge_sizes = wedge_sizes
        self.draw_legend = draw_legend

        # set colours
        if colours is not None:
            logger.info("Using specified colours list.")
            self.colours = colours
        else:
            # Using one of the colour schemes defined in puma.utils.get_good_pie_colours
            logger.info("Using specified colour scheme (%s).", colour_scheme)
            self.colours = get_good_pie_colours(colour_scheme)[: len(wedge_sizes)]

        self.labels = labels if labels is not None else ["" for i in range(len(wedge_sizes))]

        # Add some good defaults if not specified:
        self.mpl_pie_kwargs = {
            "autopct": "%1.1f%%",
            "startangle": 90,
        }
        # If mpl.pie kwargs were specified, overwrite the defaults
        if mpl_pie_kwargs is not None:
            for key, value in mpl_pie_kwargs.items():
                self.mpl_pie_kwargs[key] = value

        self.initialise_figure()
        self.plot()

    def plot(
        self,
    ):
        """Plot the pie chart."""
        self.axis_top.pie(
            x=self.wedge_sizes,
            labels=None if self.draw_legend else self.labels,
            colors=self.colours,
            **self.mpl_pie_kwargs,
        )

        # If the legend should be drawn, get the handles and plot it on the right axis
        if self.draw_legend:
            plt_handles = []
            for pie_label, pie_colour in zip(self.labels, self.colours):
                plt_handles.append(
                    mpl.patches.Patch(
                        label=pie_label,
                        color=pie_colour,
                    )
                )
            self.axis_leg.axis("off")
            self.make_legend(plt_handles, ax_mpl=self.axis_leg)
        else:
            self.axis_top.axis("equal")

        self.plotting_done = True

        if self.apply_atlas_style:
            self.atlasify()
            # Remove the legend that is automatically created by atlasify
            if not self.draw_legend:
                self.axis_top.legend().remove()

        self.set_title()
        self.set_y_lim()



"""Histogram plot functions."""

from __future__ import annotations

import matplotlib as mpl
import numpy as np
import pandas as pd
from ftag import Flavours, Label

from puma.plot_base import PlotBase, PlotLineObject
from puma.utils import get_good_colours, logger
from puma.utils.histogram import hist_ratio, hist_w_unc


class Histogram(PlotLineObject):
    """
    Histogram class storing info about histogram and allows to calculate ratio w.r.t
    other histograms.
    """

    def __init__(
        self,
        values: np.ndarray,
        weights: np.ndarray = None,
        bin_edges: np.ndarray = None,
        sum_squared_weights: np.ndarray = None,
        ratio_group: str | None = None,
        flavour: str | Label = None,
        add_flavour_label: bool = True,
        histtype: str = "step",
        is_data: bool = False,
        **kwargs,
    ) -> None:
        """Initialise properties of histogram curve object.

        Parameters
        ----------
        values : np.ndarray
            Input data for the histogram. If bin_edges is specified (not None)
            then this array is treated as the bin heights.
        weights : np.ndarray, optional
            Weights for the input data. Has to be an array of same length as the input
            data with a weight for each entry. If not specified, weight 1 will be given
            to each entry. The uncertainties are calculated as the square root of the
            squared weights (for each bin separately). By default None.
        bin_edges : np.ndarray, optional
            If specified, the histogram is considered "filled": the array given to
            values is treated as if it was the bin heights corresponding to these
            bin_edges and the "weights" input is ignored. By default None.
        sum_squared_weights : np.ndarray, optional
            Only considered if the histogram is considered filled (i.e bin_edges
            is specified). It is the sum_squared_weights per bin.
            By default None.
        ratio_group : str, optional
            Name of the ratio group this histogram is compared with. The ratio group
            allows you to compare different groups of histograms within one plot.
            By default None
        flavour: str | Label, optional
            If set, the correct colour and a label prefix will be extracted from
            `puma.utils.global_config` set for this histogram.
            Allowed values are e.g. "bjets", "cjets", "ujets", "bbjets", ...
            By default None
        add_flavour_label : bool, optional
            Set to False to suppress the automatic addition of the flavour label prefix
            in the label of the curve (i.e. "$b$-jets" in the case of b-jets).
            This is ignored if `flavour` is not set. By default True
        histtype: str, optional
            `histtype` parameter which is handed to matplotlib.hist() when plotting the
            histograms. Supported values are "bar", "barstacked", "step", "stepfilled".
            By default "step"
        is_data : bool, optional
            Decide, if the plot object will be treated as data (black dots,
            no stacking), by default False
        **kwargs : kwargs
            Keyword arguments passed to `puma.plot_base.PlotLineObject`

        Raises
        ------
        ValueError
            If input data is not of type np.ndarray or list
        ValueError
            If weights are specified but have different length as the input values
        """
        super().__init__(**kwargs)

        if isinstance(values, (np.ndarray, list, pd.core.series.Series)):
            values = np.array(values)
            if len(values) == 0:
                logger.warning("Histogram is empty.")
        else:
            raise TypeError(
                "Invalid type of histogram input data. Allowed values are "
                "numpy.ndarray, list, pandas.core.series.Series"
            )
        if weights is not None and len(values) != len(weights):
            raise ValueError("`values` and `weights` are not of same length.")

        self.values = values
        self.bin_edges = bin_edges  # Important to have this defined for any histogram
        self.sum_squared_weights = sum_squared_weights

        if bin_edges is None and sum_squared_weights is not None:
            logger.warning(
                """The Histogram has no bin edges defined and is thus
                              not considered filled. Parameter `sum_squared_weights`
                              is ignored. """
            )

        # This attribute allows to know how to handle the histogram later during
        # plotting
        self.filled = bin_edges is not None

        self.weights = weights
        self.ratio_group = ratio_group
        self.flavour = Flavours[flavour] if isinstance(flavour, str) else flavour
        self.add_flavour_label = add_flavour_label
        self.histtype = histtype
        self.is_data = is_data

        # Set histogram attributes to None. They will be defined when the histograms
        # are plotted
        self.hist = None
        self.unc = None
        self.band = None
        self.key = None

        label = kwargs["label"] if "label" in kwargs and kwargs["label"] is not None else ""
        # If flavour was specified, extract configuration from global config
        if self.flavour is not None:
            if self.flavour in Flavours:
                # Use globally defined flavour colour if not specified
                if self.colour is None:
                    self.colour = self.flavour.colour
                    logger.debug("Histogram colour was set to %s", self.colour)
                # Add globally defined flavour label if not suppressed
                if self.add_flavour_label:
                    global_flavour_label = self.flavour.label
                    self.label = f"{global_flavour_label} {label}"
                else:
                    self.label = label
                logger.debug("Histogram label was set to %s", {self.label})
            else:
                logger.warning("The flavour '%s' was not found in the global config.", self.flavour)

    def divide(self, other):
        """Calculate ratio between two class objects.

        Parameters
        ----------
        other : histogram class
            Second histogram object to calculate ratio with

        Returns
        -------
        np.ndarray
            Ratio
        np.ndarray
            Ratio error

        Raises
        ------
        ValueError
            If binning is not identical between 2 objects
        ValueError
            If hist attribute is not set for one of the two histograms
        ValueError
            If bin_edges attribute is not set for one of the two histograms
        """
        if (
            self.bin_edges is None
            or other.bin_edges is None
            or self.hist is None
            or other.hist is None
        ):
            raise ValueError(
                "Can't divide histograms since bin edges and counts are not available "
                "for both histogram. Bins are filled when they are plotted."
            )

        if not np.all(self.bin_edges == other.bin_edges):
            raise ValueError("The binning of the two given objects do not match.")

        # Bins where the reference histogram is empty/zero, are given a ratio of np.inf
        # which means that the ratio plot will not have any entry in these bins.
        ratio, ratio_unc = hist_ratio(
            numerator=self.hist,
            denominator=other.hist,
            numerator_unc=self.unc,
            step=False,
        )
        # To use the matplotlib.step() function later on, the first bin is duplicated
        ratio = np.append(np.array([ratio[0]]), ratio)
        ratio_unc = np.append(np.array([ratio_unc[0]]), ratio_unc)

        return (ratio, ratio_unc)

    def divide_data_mc(
        self,
        ref_hist: np.ndarray,
    ) -> tuple:
        """
        Similar as divide, but the second item doesn't need to be a histogram object.

        Parameters
        ----------
        ref_hist : np.ndarray
            Hist weights of the reference.
        ref_unc : np.ndarray
            Uncertainties of the reference

        Returns
        -------
        tuple
            Tuple of the ratios and ratio uncertaintes for the bins
        """
        # Bins where the reference histogram is empty/zero, are given a ratio of np.inf
        # which means that the ratio plot will not have any entry in these bins.
        ratio, ratio_unc = hist_ratio(
            numerator=self.hist,
            denominator=ref_hist,
            numerator_unc=self.unc,
            step=False,
        )
        # To use the matplotlib.step() function later on, the first bin is duplicated
        ratio = np.append(np.array([ratio[0]]), ratio)
        ratio_unc = np.append(np.array([ratio_unc[0]]), ratio_unc)

        return (ratio, ratio_unc)


class HistogramPlot(PlotBase):
    """Histogram plot class."""

    def __init__(
        self,
        bins=40,
        bins_range: tuple | None = None,
        discrete_vals: list | None = None,
        norm: bool = True,
        logy: bool = False,
        bin_width_in_ylabel: bool = False,
        underoverflow: bool = True,
        grid: bool = False,
        stacked: bool = False,
        histtype: str = "bar",
        **kwargs,
    ) -> None:
        """Histogram plot properties.

        Parameters
        ----------
        bins : int or numpy.ndarray or list, optional
            If bins is an int, it defines the number of equal-width bins in the given
            range. If bins is a sequence, it defines a monotonically increasing array
            of bin edges, including the rightmost edge, allowing for non-uniform
            bin widths (like in numpy.histogram). By default 40
        bins_range : tuple, optional
            Tuple of two floats, specifying the range for the binning. If bins_range is
            specified and bins is an integer, equal-width bins from bins_range[0] to
            bins_range[1] are used for the histogram (like in numpy.histogram).
            By default None
        discrete_vals : list, optional
            List of values if a variable only has discrete values. If discrete_vals is
            specified only the bins containing these values are plotted.
            By default None.
        norm : bool, optional
            Specify if the histograms are normalised, this means that histograms are
            divided by the total numer of counts. Therefore, the sum of the bin counts
            is equal to one, but NOT the area under the curve, which would be
            sum(bin_counts * bin_width). By default True.
        logy : bool, optional
            Set log scale on y-axis, by default False.
        bin_width_in_ylabel : bool, optional
            Specify if the bin width should be added to the ylabel, by default False
        underoverflow : bool, optional
            Option to include under- and overflow values in outermost bins, by default
            True.
        grid : bool, optional
            Set the grid for the plots, by default False
        stacked : bool, optional
            Decide, if all histograms (which are not data) are stacked, by default False
        histtype : str, optional
            If stacked is used, define the type of histogram you would like to have,
            default is "bar"
        **kwargs : kwargs
            Keyword arguments from `puma.PlotObject`

        Raises
        ------
        ValueError
            If n_ratio_panels > 1
        """
        super().__init__(grid=grid, **kwargs)
        self.logy = logy
        self.bins = bins
        self.bins_range = bins_range
        self.discrete_vals = discrete_vals
        self.bin_width_in_ylabel = bin_width_in_ylabel
        self.norm = norm
        self.underoverflow = underoverflow
        self.stacked = stacked
        self.histtype = histtype
        self.plot_objects = {}
        self.add_order = []
        self.ratios_objects = {}
        self.reference_object = None

        if self.norm is True and self.stacked is True:
            raise ValueError(
                "Stacked plots and normalised plots at the same time are not available."
            )

        if self.n_ratio_panels > 1:
            raise ValueError("Not more than one ratio panel supported.")
        self.initialise_figure()

    def add(
        self,
        histogram: Histogram,
        key: str | None = None,
        reference: bool = False,
    ):
        """Adding histogram object to figure.

        Parameters
        ----------
        histogram : Histogram class
            Histogram curve
        key : str, optional
            Unique identifier for histogram, by default None
        reference : bool, optional
            If this histogram is used as reference for ratio calculation, by default
            False

        Raises
        ------
        KeyError
            If unique identifier key is used twice
        """
        if key is None:
            key = len(self.plot_objects) + 1
        if key in self.plot_objects:
            raise KeyError(f"Duplicated key {key} already used for unique identifier.")

        # Add key to histogram object
        histogram.key = key
        logger.debug("Adding histogram %s", key)

        # Set linestyle
        if histogram.linestyle is None:
            if histogram.is_data is True:
                histogram.linestyle = ""
            else:
                histogram.linestyle = "-"
        # Set marker
        if histogram.marker is None:
            if histogram.is_data is True:
                histogram.marker = "."
            else:
                histogram.marker = ""
        # Set colours
        if histogram.colour is None:
            histogram.colour = get_good_colours()[len(self.plot_objects)]
        # Set alpha
        if histogram.alpha is None:
            histogram.alpha = 1
        # Set linewidth
        if histogram.linewidth is None:
            histogram.linewidth = 1.6
        # Set markersize
        if histogram.markersize is None:
            histogram.markersize = 10

        self.plot_objects[key] = histogram
        self.add_order.append(key)
        if reference is True:
            self.set_reference(key)

    def set_reference(self, key: str):
        """Setting the reference histogram curves used in the ratios.

        Parameters
        ----------
        key : str
            Unique identifier of histogram object
        """
        if self.reference_object is None:
            self.reference_object = [key]
        else:
            self.reference_object.append(key)
        logger.debug("Adding '%s' to reference histogram(s)", key)

    def plot(self, **kwargs):
        """Plotting curves. This also generates the bins of the histograms that are
        added to the plot. Plot objects are drawn in the same order as they were added
        to the plot.

        Parameters
        ----------
        **kwargs: kwargs
            Keyword arguments passed to matplotlib.axes.Axes.hist()

        Returns
        -------
        Line2D
            matplotlib Line2D object

        Raises
        ------
        ValueError
            If specified bins type is not supported.
        """
        if self.ylabel is not None and self.norm and "norm" not in self.ylabel.lower():
            logger.warning(
                "You are plotting normalised distributions but 'norm' is not "
                "included in your y-label."
            )
        plt_handles = []

        # Calculate bins of stacked histograms to ensure all histograms fit in plot
        if isinstance(self.bins, (np.ndarray, list)):
            logger.debug("Using bin edges defined in plot instance.")
            if self.bins_range is not None:
                logger.warning(
                    "You defined a range for the histogram, but also an array with "
                    "the bin edges. The range will be ignored."
                )
        elif isinstance(self.bins, int):
            logger.debug("Calculating bin edges of %i equal-width bins", self.bins)
            _, self.bins = np.histogram(
                np.hstack([elem.values for elem in self.plot_objects.values()]),  # noqa: PD011
                bins=self.bins,
                range=self.bins_range,
            )
        else:
            raise TypeError("Unsupported type for bins. Supported types: int, numpy.array, list")

        # Loop over all plot objects and plot them
        bins = self.bins

        # Stacked dict for the stacked histogram
        self.stacked_dict = {
            "x": [],
            "weights": [],
            "color": [],
            "unc": None,
        }

        for key in self.add_order:
            elem = self.plot_objects[key]

            elem.bin_edges, elem.hist, elem.unc, elem.band = hist_w_unc(
                elem.values,
                weights=elem.weights,
                bin_edges=elem.bin_edges,
                sum_squared_weights=elem.sum_squared_weights,
                bins=self.bins,
                filled=elem.filled,
                bins_range=self.bins_range,
                normed=self.norm,
                underoverflow=self.underoverflow,
            )

            # MAYBE CHECK HERE THAT self.bins and elem.bin_edges are
            # equivalent for plotting or throw error!

            if self.discrete_vals is not None:
                # bins are recalculated for the discrete values
                bins = self.get_discrete_values(elem)

            # Check if the histogram is data
            if elem.is_data is True:
                # Plot data
                self.axis_top.errorbar(
                    x=(bins[:-1] + bins[1:]) / 2,
                    y=elem.hist,
                    yerr=elem.unc if self.draw_errors else 0,
                    color=elem.colour,
                    label=elem.label,
                    alpha=elem.alpha,
                    linewidth=elem.linewidth,
                    linestyle=elem.linestyle,
                    marker=elem.marker,
                    markersize=elem.markersize,
                )

                plt_handles.append(
                    mpl.lines.Line2D(
                        [],
                        [],
                        color=elem.colour,
                        label=elem.label,
                        alpha=elem.alpha,
                        linewidth=elem.linewidth,
                        linestyle=elem.linestyle,
                        marker=elem.marker,
                    )
                )

            elif self.stacked:
                self.stacked_dict["x"].append(bins[:-1])
                self.stacked_dict["weights"].append(elem.hist)
                self.stacked_dict["color"].append(elem.colour)

                if self.stacked_dict["unc"] is None:
                    self.stacked_dict["unc"] = elem.unc

                else:
                    self.stacked_dict["unc"] = np.sqrt(self.stacked_dict["unc"] ** 2 + elem.unc**2)

                # Add the element to the legend with a "bar"
                plt_handles.append(
                    mpl.patches.Patch(
                        color=elem.colour,
                        label=elem.label,
                        alpha=elem.alpha,
                    )
                )

            else:
                # Plot histogram
                self.axis_top.hist(
                    x=bins[:-1],
                    bins=bins,
                    weights=elem.hist,
                    histtype=elem.histtype,
                    color=elem.colour,
                    label=elem.label,
                    alpha=elem.alpha,
                    linewidth=elem.linewidth,
                    linestyle=elem.linestyle,
                    **kwargs,
                )

                # Plot histogram uncertainty
                if self.draw_errors:
                    bottom_error = np.array([elem.band[0], *elem.band.tolist()])
                    top_error = elem.band + 2 * elem.unc
                    top_error = np.array([top_error[0], *top_error.tolist()])
                    self.axis_top.fill_between(
                        x=elem.bin_edges,
                        y1=bottom_error,
                        y2=top_error,
                        color=elem.colour,
                        alpha=0.3,
                        zorder=1,
                        step="pre",
                        edgecolor="none",
                    )

                # Add standard "Line" to legend
                plt_handles.append(
                    mpl.lines.Line2D(
                        [],
                        [],
                        color=elem.colour,
                        label=elem.label,
                        alpha=elem.alpha,
                        linewidth=elem.linewidth,
                        linestyle=elem.linestyle,
                        marker=elem.marker,
                    )
                )

        if self.stacked:
            self.axis_top.hist(
                x=self.stacked_dict["x"],
                bins=bins,
                weights=self.stacked_dict["weights"],
                color=self.stacked_dict["color"],
                histtype=self.histtype,
                alpha=elem.alpha,
                linewidth=elem.linewidth,
                linestyle=elem.linestyle,
                stacked=self.stacked,
                **kwargs,
            )

        # Check if errors should be drawn
        # If stacked is true, plot the combined uncertainty
        if self.draw_errors and self.stacked:
            # Create a total weights entry to correctly plot the error band
            # Total weights is here the y-value of all contributions stacked
            self.stacked_dict["total_weights"] = np.sum(self.stacked_dict["weights"], axis=0)

            # Calculate the y-values of the bottom error
            bottom_error = self.stacked_dict["total_weights"] - self.stacked_dict["unc"]
            bottom_error = np.array([bottom_error[0], *bottom_error.tolist()])

            # Calculate the y-values of the top error
            top_error = self.stacked_dict["total_weights"] + self.stacked_dict["unc"]
            top_error = np.array([top_error[0], *top_error.tolist()])

            # Fill the space between bottom and top with the unc. band
            self.axis_top.fill_between(
                x=elem.bin_edges,
                y1=bottom_error,
                y2=top_error,
                alpha=0.5,
                zorder=1,
                step="pre",
                facecolor="white",
                edgecolor="black",
                linewidth=0,
                hatch="/////",
            )

            # Add a label for the unc. in the legend
            plt_handles.append(
                mpl.patches.Patch(
                    facecolor="white",
                    edgecolor="black",
                    label="Stat. unc.",
                    linewidth=0,
                    alpha=0.5,
                    hatch="/////",
                )
            )

        if self.discrete_vals is not None:
            self.bins = bins

        self.plotting_done = True
        return plt_handles

    def get_discrete_values(self, elem: object):
        """Get discrete values of a variable and adjust the
        bins accordingly.

        Parameters
        ----------
        elem : histogram class
            Histogram we want to calculate the bins containing discrete values for

        Returns
        -------
        bins : numpy.ndarray
            Recalculated bins including only the discrete values

        Raises
        ------
        ValueError
            If the bin width is larger than 1 such that potentially not
            all discrete values are in a seperate bin
        ValueError
            If the number of bins is set to 1 such that no values can be
            distinguished
        """
        if len(elem.bin_edges) > 1:
            if abs(elem.bin_edges[1] - elem.bin_edges[0]) <= 1:
                indice = [
                    i
                    for i in range(len(elem.bin_edges) - 1)
                    for discrete_val in self.discrete_vals
                    if elem.bin_edges[i] <= discrete_val < elem.bin_edges[i + 1]
                ]
                elem.hist = elem.hist[indice]
                elem.unc = elem.unc[indice]
                elem.band = elem.band[indice]
                bins = np.linspace(0, len(self.discrete_vals), len(self.discrete_vals) + 1)
                elem.bin_edges = bins
                self.axis_top.set_xticks(bins[:-1] + 0.5)
                self.axis_top.set_xticklabels(self.discrete_vals, rotation=33)
            else:
                raise ValueError(
                    "Bin width is larger than 1. Choose a binning with a bin"
                    " width<= 1 to plot only discrete values."
                )
        else:
            raise ValueError(
                "Choose a binning with more than one bin in order to plot only discrete" " values."
            )

        return bins

    def get_reference_histo(self, histo):
        """Get reference histogram from list of references.

        Parameters
        ----------
        histo : puma.histogram.Histogram
            Histogram we want to calculate the ratio for

        Returns
        -------
        reference_histo_name : str, int
            Identifier of the corresponding reference histogram

        Raises
        ------
        ValueError
            If no reference histo was found or multiple matches.
        """
        matches = 0
        reference_histo = None

        for key in self.reference_object:
            reference_candidate = self.plot_objects[key]
            if histo.ratio_group is not None:
                if histo.ratio_group == reference_candidate.ratio_group:
                    matches += 1
                    reference_histo = reference_candidate
            else:
                matches += 1
                reference_histo = reference_candidate

        if matches != 1:
            raise ValueError(
                f"Found {matches} matching reference candidates, but only one match is" " allowed."
            )

        logger.debug("Reference histogram for '%s' is '%s'", histo.key, reference_histo.key)

        return reference_histo

    def plot_ratios(self):
        """Plotting ratio histograms.

        Raises
        ------
        ValueError
            If no reference histogram is defined
        """
        # Check if this is a stacked plot
        # Plot ratio only between data and the stacked histos
        for key in self.add_order:
            # Get the object which is to be plotted
            elem = self.plot_objects[key]

            if elem.bin_edges is None:
                raise ValueError(
                    "Bin edges of plot object not set. This is done in "
                    "histogram_plot.plot(), so it has to be called before "
                    "plot_ratios() is called."
                )

            # Check if this is going to be Data/MC (Data/stacked plot)
            if self.stacked:
                # Check this is data
                if not elem.is_data:
                    continue

                # Using the total weights (full stacked histo) as reference for data
                ratio, ratio_unc = elem.divide_data_mc(
                    ref_hist=self.stacked_dict["total_weights"],
                )

            else:
                if self.reference_object is None:
                    raise ValueError("Please specify a reference curve.")

                ratio, ratio_unc = elem.divide(self.get_reference_histo(elem))

            # Plot the ratio values with the step function
            if self.stacked:
                if elem.is_data is True:
                    self.ratio_axes[0].errorbar(
                        x=(elem.bin_edges[:-1] + elem.bin_edges[1:]) / 2,
                        y=ratio[1:],
                        yerr=ratio_unc[1:] if self.draw_errors else 0,
                        color=elem.colour,
                        label=elem.label,
                        alpha=elem.alpha,
                        linewidth=elem.linewidth,
                        linestyle=elem.linestyle,
                        marker=elem.marker,
                        markersize=elem.markersize,
                    )

            else:
                self.ratio_axes[0].step(
                    x=elem.bin_edges,
                    y=ratio,
                    color=elem.colour,
                    linewidth=elem.linewidth,
                    linestyle=elem.linestyle,
                )

                # Plot the ratio uncertainty
                if self.draw_errors:
                    self.ratio_axes[0].fill_between(
                        x=elem.bin_edges,
                        y1=np.nan_to_num(ratio - ratio_unc, nan=0, posinf=0),
                        y2=np.nan_to_num(ratio + ratio_unc, nan=0, posinf=0),
                        color=elem.colour,
                        alpha=0.3,
                        zorder=1,
                        step="pre",
                        edgecolor="none",
                    )

        if self.stacked and self.draw_errors:
            self.ratio_axes[0].fill_between(
                x=elem.bin_edges,
                y1=np.nan_to_num((ratio - ratio_unc) / ratio, nan=0, posinf=0),
                y2=np.nan_to_num((ratio + ratio_unc) / ratio, nan=0, posinf=0),
                color=elem.colour,
                alpha=0.3,
                zorder=1,
                step="pre",
                edgecolor="none",
            )

    def add_bin_width_to_ylabel(self):
        """Adds the bin width to the ylabel of a histogram plot. If the bin with is
        smaller than 0.01, scientific notation will be used.

        Raises
        ------
        ValueError
            If plotting_done is False (therefore `bins` is not yet calculated)
        """
        if self.plotting_done is False:
            raise ValueError(
                "`add_bin_width_to_ylabel` should be called after plotting, since bins "
                "are calculated during plotting."
            )

        bin_width = abs(self.bins[1] - self.bins[0])
        if bin_width < 1e-2:
            self.ylabel = f"{self.ylabel} / {bin_width:.0e}"
        else:
            self.ylabel = f"{self.ylabel} / {bin_width:.2f}"
        self.set_ylabel(self.axis_top)

    def draw(self, labelpad: int | None = None):
        """Draw figure.

        Parameters
        ----------
        labelpad : int, optional
            Spacing in points from the axes bounding box including
            ticks and tick labels, by default "ratio"

        """
        plt_handles = self.plot()

        if self.n_ratio_panels > 0:
            self.plot_ratios()

        self.set_xlim(
            self.bins[0] if self.xmin is None else self.xmin,
            self.bins[-1] if self.xmax is None else self.xmax,
        )
        self.set_log()
        self.set_y_lim()
        self.set_xlabel()
        self.set_tick_params()
        self.set_ylabel(self.axis_top)

        if self.n_ratio_panels > 0:
            self.set_ylabel(
                self.ratio_axes[0],
                self.ylabel_ratio[0],
                align="center",
                labelpad=labelpad,
            )

        if self.bin_width_in_ylabel is True:
            self.add_bin_width_to_ylabel()

        legend_axis = self.axis_top

        self.make_legend(plt_handles, ax_mpl=legend_axis)
        self.set_title()

        if self.apply_atlas_style:
            self.atlasify()



"""Unit test script for the functions in roc.py."""

from __future__ import annotations

import os
import shutil  # noqa: F401
import tempfile
import unittest

import numpy as np
from ftag import get_discriminant
from matplotlib.testing.compare import compare_images

from puma import IntegratedEfficiency, IntegratedEfficiencyPlot
from puma.utils import get_dummy_2_taggers
from puma.utils.logging import logger, set_log_level

set_log_level(logger, "DEBUG")


class IntegratedEfficiencyCase(unittest.TestCase):
    """Test class for the puma.roc functions."""

    def setUp(self):
        self.disc_sig = np.random.normal(0.7, 1, 100)
        self.disc_rej = np.random.normal(0.4, 1, 100)

    def test_int_eff_init(self):
        """Test init."""
        IntegratedEfficiency(self.disc_sig, self.disc_rej)

    def test_add_label_flavour(self):
        """Test both label and flavour."""
        int_eff = IntegratedEfficiency(
            self.disc_sig, self.disc_rej, label="b-jets", flavour="ujets"
        )
        self.assertEqual(int_eff.label, "b-jets")


class IntegratedEfficiencyPlotTestCase(unittest.TestCase):
    """Test class for the puma.IntegratedEfficiencyPlot."""

    def setUp(self):
        # Set up temp directory for comparison plots
        self.tmp_dir = tempfile.TemporaryDirectory()  # pylint: disable=R1732
        self.actual_plots_dir = f"{self.tmp_dir.name}/"
        self.expected_plots_dir = os.path.join(os.path.dirname(__file__), "expected_plots")

        # Set up dummy data
        df = get_dummy_2_taggers(size=int(1e6))
        is_light = df["HadronConeExclTruthLabelID"] == 0
        is_c = df["HadronConeExclTruthLabelID"] == 4
        is_b = df["HadronConeExclTruthLabelID"] == 5

        disc_dips = get_discriminant(df, "dips", "bjets", fc=0.018)
        disc_rnnip = get_discriminant(df, "rnnip", "bjets", fc=0.018)

        self.dips_int_effs = {
            "light": IntegratedEfficiency(
                disc_dips[is_b],
                disc_dips[is_light],
                n_vals=200,
                tagger="DIPS",
                flavour="ujets",
            ),
            "c": IntegratedEfficiency(
                disc_dips[is_b],
                disc_dips[is_c],
                n_vals=200,
                tagger="DIPS",
                flavour="cjets",
            ),
            "b": IntegratedEfficiency(
                disc_dips[is_b],
                disc_dips[is_b],
                n_vals=200,
                tagger="DIPS",
                flavour="bjets",
            ),
        }
        self.rnnip_int_effs = {
            "light": IntegratedEfficiency(
                disc_rnnip[is_b],
                disc_rnnip[is_light],
                n_vals=200,
                tagger="RNNIP",
                flavour="ujets",
            ),
            "c": IntegratedEfficiency(
                disc_rnnip[is_b],
                disc_rnnip[is_c],
                n_vals=200,
                tagger="RNNIP",
                flavour="cjets",
            ),
            "b": IntegratedEfficiency(
                disc_rnnip[is_b],
                disc_rnnip[is_b],
                n_vals=200,
                tagger="RNNIP",
                flavour="bjets",
            ),
        }

    def test_duplicate_key(self):
        """Test duplicate key."""
        plot = IntegratedEfficiencyPlot()
        plot.add(self.dips_int_effs["b"], key=1)
        with self.assertRaises(KeyError):
            plot.add(self.dips_int_effs["c"], key=1)

    def test_output_one_tagger(self):
        """Test with one tagger."""
        plot = IntegratedEfficiencyPlot(grid=True)
        for flav in ["b", "c", "light"]:
            plot.add(self.dips_int_effs[flav])
        plot.draw()
        name = "test_int_eff_one_tagger.png"
        plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_output_two_taggers(self):
        """Test with one tagger."""
        plot = IntegratedEfficiencyPlot(grid=True)
        for flav in ["b", "c", "light"]:
            plot.add(self.dips_int_effs[flav])
            plot.add(self.rnnip_int_effs[flav])
        plot.draw()
        name = "test_int_eff_two_taggers.png"
        plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )



"""
Install the package and all its dependencies.

Run

    python -m pip install .

to install the package.
"""

# read the contents of the README.md file
from __future__ import annotations

from pathlib import Path

from setuptools import setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    long_description=long_description,
    long_description_content_type="text/markdown",
)



"""puma framework - Plotting UMami Api."""

# flake8: noqa

__version__ = "0.4.3"

from puma.histogram import Histogram, HistogramPlot
from puma.integrated_eff import IntegratedEfficiency, IntegratedEfficiencyPlot
from puma.line_plot_2d import Line2D, Line2DPlot
from puma.pie import PiePlot
from puma.plot_base import PlotBase, PlotLineObject, PlotObject
from puma.roc import Roc, RocPlot
from puma.var_vs_eff import VarVsEff, VarVsEffPlot
from puma.var_vs_var import VarVsVar, VarVsVarPlot
from puma.var_vs_vtx import VarVsVtx, VarVsVtxPlot

__all__ = [
    "Histogram",
    "HistogramPlot",
    "Line2D",
    "Line2DPlot",
    "PiePlot",
    "PlotBase",
    "PlotLineObject",
    "PlotObject",
    "Roc",
    "RocPlot",
    "VarVsEff",
    "VarVsEffPlot",
    "VarVsVar",
    "VarVsVarPlot",
    "VarVsVtx",
    "VarVsVtxPlot",
    "IntegratedEfficiency",
    "IntegratedEfficiencyPlot",
]



"""Unit test script for the functions in histogram.py."""

from __future__ import annotations

import os
import shutil  # noqa: F401
import tempfile
import unittest

import numpy as np
from ftag import Flavours
from matplotlib.testing.compare import compare_images

from puma import Histogram, HistogramPlot
from puma.utils import logger, set_log_level

set_log_level(logger, "DEBUG")


class HistogramTestCase(unittest.TestCase):
    """Test class for the puma.histogram functions."""

    def test_empty_histogram(self):
        """Test if providing wrong input type to histogram raises ValueError."""
        with self.assertRaises(TypeError):
            Histogram(values=5)

    def test_divide_before_plotting(self):
        """Test if ValueError is raised when dividing before plotting the histograms."""
        hist_1 = Histogram([1, 1, 1, 2, 2])
        hist_2 = Histogram([1, 2, 2, 2])
        with self.assertRaises(ValueError):
            hist_1.divide(hist_2)

    def test_divide_after_plotting_no_norm(self):
        """Test if ratio is calculated correctly after plotting (without norm)."""
        hist_1 = Histogram([1, 1, 1, 2, 2])
        hist_2 = Histogram([1, 2, 2, 2])
        bins = np.array([1, 2, 3])
        hist_plot = HistogramPlot(bins=bins, norm=False)
        hist_plot.add(hist_1)
        hist_plot.add(hist_2)
        hist_plot.plot()
        # Since plotting is done with the matplotlib step function, the first bin is
        # duplicated in the ratio calculation (the first one is so to say not plotted)
        # Therefore, we also use duplicated bins here
        expected_ratio = np.array([3, 3, 2 / 3])
        expected_ratio_unc = np.array([1.73205081, 1.73205081, 0.47140452])

        np.testing.assert_almost_equal(expected_ratio, hist_1.divide(hist_2)[0])
        np.testing.assert_almost_equal(expected_ratio_unc, hist_1.divide(hist_2)[1])

    def test_divide_after_plotting_norm(self):
        """Test if ratio is calculated correctly after plotting (with norm)."""
        hist_1 = Histogram([1, 1, 1, 2, 2])
        hist_2 = Histogram([1, 2, 2, 2])
        bins = np.array([1, 2, 3])
        hist_plot = HistogramPlot(bins=bins, norm=True)
        hist_plot.add(hist_1)
        hist_plot.add(hist_2)
        hist_plot.plot()
        # Since plotting is done with the matplotlib step function, the first bin is
        # duplicated in the ratio calculation (the first one is so to say not plotted)
        # Therefore, we also use duplicated bins here
        expected_ratio = np.array([2.4, 2.4, 0.53333333])
        expected_ratio_unc = np.array([1.38564065, 1.38564065, 0.37712362])

        np.testing.assert_almost_equal(expected_ratio, hist_1.divide(hist_2)[0])
        np.testing.assert_almost_equal(expected_ratio_unc, hist_1.divide(hist_2)[1])

    def test_ratio_same_histogram(self):
        """Test if ratio is 1 for equal histograms (with norm)."""
        hist_1 = Histogram([1, 1, 1, 2, 2])
        hist_2 = Histogram([1, 1, 1, 2, 2])
        bins = np.array([1, 2, 3])
        hist_plot = HistogramPlot(bins=bins, norm=True)
        hist_plot.add(hist_1)
        hist_plot.add(hist_2)
        hist_plot.plot()
        # Since plotting is done with the matplotlib step function, the first bin is
        # duplicated in the ratio calculation (the first one is so to say not plotted)
        # Therefore, we also use duplicated bins here
        expected_ratio = np.ones(3)
        expected_ratio_unc = np.array([0.57735027, 0.57735027, 0.70710678])

        np.testing.assert_almost_equal(expected_ratio, hist_1.divide(hist_2)[0])
        np.testing.assert_almost_equal(expected_ratio_unc, hist_1.divide(hist_2)[1])

    def test_divide_wrong_bin_edges(self):
        """Test if error is raised if bin edges don't match."""
        hist_1 = Histogram([1, 1, 1, 2, 2])
        hist_2 = Histogram([1, 2, 2, 2])
        bins = np.array([1, 2, 3])
        hist_plot = HistogramPlot(bins=bins, norm=False)
        hist_plot.add(hist_1)
        hist_plot.add(hist_2)
        hist_plot.plot()
        # Since plotting is done with the matplotlib step function, the first bin is
        # duplicated in the ratio calculation (the first one is so to say not plotted)
        # Therefore, we also use duplicated bins here
        hist_2.bin_edges = np.array([1, 2])
        with self.assertRaises(ValueError):
            hist_1.divide(hist_2)

    def test_multiple_references_wrong_flavour(self):
        """Tests if warning is raised with wrong flavour."""
        dummy_array = np.array([1, 1, 2, 3, 2, 3])
        with self.assertRaises(KeyError):
            Histogram(dummy_array, flavour="dummy")


class HistogramPlotTestCase(unittest.TestCase):
    """Test class for puma.histogram_plot."""

    def setUp(self):
        np.random.seed(42)
        n_random = 10_000
        self.hist_1 = Histogram(np.random.normal(size=n_random), label=f"N={n_random:_}")
        self.hist_2 = Histogram(np.random.normal(size=2 * n_random), label=f"N={2 * n_random:_}")
        self.data_hist = Histogram(
            np.random.normal(size=3 * n_random),
            label=f"Toy Data, N={3 * n_random:_}",
            is_data=True,
            colour="k",
        )

        # Set up directories for comparison plots
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.actual_plots_dir = f"{self.tmp_dir.name}/"
        self.expected_plots_dir = os.path.join(os.path.dirname(__file__), "expected_plots")

    def test_invalid_bins_type(self):
        """Check if ValueError is raised when using invalid type in `bins` argument."""
        hist_plot = HistogramPlot(bins=1.1)
        hist_plot.add(self.hist_1, reference=True)
        with self.assertRaises(TypeError):
            hist_plot.plot()

    def test_add_bin_width_to_ylabel(self):
        """Check if ValueError is raised when using invalid type in `bins` argument."""
        hist_plot = HistogramPlot(bins=60)
        hist_plot.add(self.hist_1, reference=True)
        with self.assertRaises(ValueError):
            hist_plot.add_bin_width_to_ylabel()

    def test_multiple_references_no_flavour(self):
        """Tests if error is raised in case of non-unique reference histogram."""
        hist_plot = HistogramPlot(n_ratio_panels=1)
        hist_plot.add(self.hist_1, reference=True)
        hist_plot.add(self.hist_2, reference=True)
        hist_plot.add(self.hist_1)
        hist_plot.plot()
        with self.assertRaises(ValueError):
            hist_plot.plot_ratios()

    def test_multiple_references_flavour(self):
        """Tests if error is raised in case of non-unique reference histogram
        when flavoured histograms are used
        .
        """
        hist_plot = HistogramPlot(n_ratio_panels=1)
        dummy_array = np.array([1, 1, 2, 3, 2, 3])
        hist_plot.add(Histogram(dummy_array, flavour="ujets"), reference=True)
        hist_plot.add(Histogram(dummy_array, flavour="ujets"), reference=True)
        hist_plot.add(Histogram(dummy_array, flavour="ujets"))
        hist_plot.plot()
        with self.assertRaises(ValueError):
            hist_plot.plot_ratios()

    def test_two_ratio_panels(self):
        """Tests if error is raised if more than 1 ratio panel is requested."""
        with self.assertRaises(ValueError):
            HistogramPlot(n_ratio_panels=2)

    def test_custom_range(self):
        """Check if
        1. bins_range argument is used correctly
        2. deactivate ATLAS branding works
        3. adding bin width to ylabel works.
        """
        hist_plot = HistogramPlot(
            bins=20,
            bins_range=(0, 4),
            atlas_brand="",
            atlas_first_tag="Simulation, $\\sqrt{s}=13$ TeV",
            atlas_second_tag="Second tag with additional\ndistance from first tag",
            figsize=(5, 4),
            ylabel="Number of jets",
            n_ratio_panels=1,
            atlas_second_tag_distance=0.3,
        )
        hist_plot.add(self.hist_1, reference=True)
        hist_plot.draw()
        hist_plot.add_bin_width_to_ylabel()

        name = "test_histogram_custom_range.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_data_mc(self):
        """Check if data mc looks good."""
        hist_plot = HistogramPlot(
            bins=20,
            bins_range=[-2, 2],
            atlas_brand="",
            atlas_first_tag="Simulation, $\\sqrt{s}=13$ TeV",
            atlas_second_tag="Second tag with additional\ndistance from first tag",
            figsize=(5, 4),
            ylabel="Number of jets",
            n_ratio_panels=1,
            atlas_second_tag_distance=0.3,
            stacked=True,
            norm=False,
        )
        hist_plot.add(self.hist_1)
        hist_plot.add(self.hist_2)
        hist_plot.add(self.data_hist)
        hist_plot.draw()
        hist_plot.add_bin_width_to_ylabel()

        name = "test_histogram_data_mc.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_discrete_values(self):
        """Check if discrete values are working properly."""
        hist_plot = HistogramPlot(
            bins=np.linspace(0, 10, 100),
            discrete_vals=[0, 5, 7, 9],
            atlas_first_tag="Simulation, $\\sqrt{s}=13$ TeV",
            figsize=(5, 4),
            xlabel="Discrete values",
            ylabel="Number of counts in specified bins",
            n_ratio_panels=1,
        )
        # the entry "1" in `values_1` will be hidden in the histogram since it is not
        # included in the `discrete_vals` list
        values1 = np.array([0, 1, 5, 7])
        values2 = np.array([0, 5, 5, 7])
        hist_plot.add(Histogram(values1), reference=True)
        hist_plot.add(Histogram(values2, linestyle="--"))
        hist_plot.draw()

        name = "test_histogram_discrete_values.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_output_ratio(self):
        """Check with a plot if the ratio is the expected value."""
        hist_plot = HistogramPlot(
            norm=False,
            ymax_ratio=[4],
            figsize=(6.5, 5),
            n_ratio_panels=1,
        )
        hist_plot.add(self.hist_1, reference=True)
        hist_plot.add(self.hist_2)
        hist_plot.draw()
        hist_plot.ratio_axes[0].axhline(2, color="r", label="Expected ratio")
        hist_plot.ratio_axes[0].legend(frameon=False)

        name = "test_histogram_ratio_value.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

        # Also save this plot with transparent background to test this feature
        name_transparent = "test_histogram_ratio_value_transparent.png"
        hist_plot.transparent = True
        hist_plot.savefig(f"{self.actual_plots_dir}/{name_transparent}")
        # Uncomment line below to update expected image
        # shutil.copy(
        #     f"{self.actual_plots_dir}/{name_transparent}",
        #     f"{self.expected_plots_dir}/{name_transparent}",
        # )
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name_transparent}",
                f"{self.expected_plots_dir}/{name_transparent}",
                tol=1,
            )
        )

    def test_output_empty_histogram_norm(self):
        """Test empty normed histogram."""
        hist_plot = HistogramPlot(
            norm=True,
            figsize=(6.5, 5),
            atlas_second_tag=(
                "Test if ratio is 1 for whole range if reference histogram is empty\n(+"
                " normalised)"
            ),
            n_ratio_panels=1,
        )
        hist_plot.add(Histogram(np.array([]), label="empty histogram"), reference=True)
        hist_plot.add(self.hist_1)
        hist_plot.draw()

        name = "test_histogram_empty_reference_norm.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_output_empty_histogram_no_norm(self):
        """Test if ratio is 1 for whole range if reference histogram is empty."""
        hist_plot = HistogramPlot(
            norm=False,
            figsize=(6.5, 5),
            atlas_second_tag=("Test if ratio is 1 for whole range if reference histogram is empty"),
            n_ratio_panels=1,
        )
        hist_plot.add(Histogram(np.array([]), label="empty histogram"), reference=True)
        hist_plot.add(self.hist_1)
        hist_plot.draw()

        name = "test_histogram_empty_reference_no_norm.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_output_different_range_histogram(self):
        """Test if ratio yields the expected values for case of different histogram
        ranges
        .
        """
        hist_plot = HistogramPlot(
            atlas_second_tag=("Test ratio for the case of different histogram ranges. \n"),
            xlabel="x",
            figsize=(7, 6),
            leg_loc="upper right",
            y_scale=1.5,
            n_ratio_panels=1,
        )
        np.random.seed(42)
        n_random = 10_000
        arr_1 = np.concatenate((
            np.random.uniform(-2, 0, n_random),
            np.random.uniform(0.5, 0.99, int(0.5 * n_random)),
        ))
        arr_2 = np.random.uniform(0, 2, n_random)
        arr_3 = np.random.uniform(-1, 1, n_random)
        hist_plot.add(
            Histogram(arr_1, label="uniform [-2, 0] and uniform [0.5, 1] \n(reference)"),
            reference=True,
        )
        hist_plot.add(Histogram(arr_2, label="uniform [0, 2]"))
        hist_plot.add(Histogram(arr_3, label="uniform [-1, 1]"))
        hist_plot.draw()

        name = "test_histogram_different_ranges.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_draw_vlines_histogram(self):
        """Test the drawing of vertical lines in the histogram."""
        hist_plot = HistogramPlot(
            bins=20,
            bins_range=(0, 4),
            atlas_brand="",
            atlas_first_tag="Simulation, $\\sqrt{s}=13$ TeV",
            atlas_tag_outside=True,
            figsize=(5, 4),
            ylabel="Number of jets",
            n_ratio_panels=1,
        )
        hist_plot.add(self.hist_1, reference=True)
        hist_plot.draw_vlines(
            xs=[1, 2, 3],
        )
        hist_plot.draw()

        name = "test_draw_vlines_histogram.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_draw_vlines_histogram_custom_labels(self):
        """Test the drawing of vertical lines in the histogram."""
        hist_plot = HistogramPlot(
            bins=20,
            bins_range=(0, 4),
            atlas_brand="",
            atlas_first_tag="Simulation, $\\sqrt{s}=13$ TeV",
            atlas_second_tag="",
            figsize=(5, 4),
            ylabel="Number of jets",
            n_ratio_panels=1,
        )
        hist_plot.add(self.hist_1, reference=True)
        hist_plot.draw_vlines(
            xs=[1, 2, 3],
            labels=["One", "Two", "Three"],
        )
        hist_plot.draw()

        name = "test_draw_vlines_histogram_custom_labels.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_draw_vlines_histogram_same_height(self):
        """Test the drawing of vertical lines in the histogram."""
        hist_plot = HistogramPlot(
            bins=20,
            bins_range=(0, 4),
            atlas_brand="",
            atlas_first_tag="Simulation, $\\sqrt{s}=13$ TeV",
            atlas_second_tag="",
            figsize=(5, 4),
            ylabel="Number of jets",
            n_ratio_panels=1,
        )
        hist_plot.add(self.hist_1, reference=True)
        hist_plot.draw_vlines(
            xs=[1, 2, 3],
            same_height=True,
        )
        hist_plot.draw()

        name = "test_draw_vlines_histogram_same_height.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_draw_vlines_histogram_custom_yheight(self):
        """Test the drawing of vertical lines in the histogram."""
        hist_plot = HistogramPlot(
            bins=20,
            bins_range=(0, 4),
            atlas_brand="",
            atlas_first_tag="Simulation, $\\sqrt{s}=13$ TeV",
            atlas_second_tag="",
            figsize=(5, 4),
            ylabel="Number of jets",
            n_ratio_panels=1,
        )
        hist_plot.add(self.hist_1, reference=True)
        hist_plot.draw_vlines(
            xs=[1, 2, 3],
            ys=[0.7, 0.6, 0.5],
        )
        hist_plot.draw()

        name = "test_draw_vlines_histogram_custom_yheight.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_ratio_group_argument(self):
        """Test different combinations of using ratio groups."""
        hist_plot = HistogramPlot(
            n_ratio_panels=1,
            atlas_brand=None,
            atlas_first_tag="",
            atlas_second_tag=(
                "Unit test plot to test the ratio_group argument \n"
                "of the puma.Histogram class.\n"
                "Also testing the linestyle legend which is\n"
                "placed here at (0.25, 0.75)."
            ),
            figsize=(6, 5),
            y_scale=1.5,
        )

        rng = np.random.default_rng(seed=42)
        # add two histograms with flavour=None but ratio_goup set
        hist_plot.add(
            Histogram(
                rng.normal(0, 1, size=10_000),
                ratio_group="ratio group 1",
                colour=Flavours["ujets"].colour,
                label=Flavours["ujets"].label,
            ),
            reference=True,
        )
        hist_plot.add(
            Histogram(
                rng.normal(0.5, 1, size=10_000),
                colour=Flavours["ujets"].colour,
                ratio_group="ratio group 1",
                linestyle="--",
            ),
        )
        hist_plot.add(
            Histogram(
                rng.normal(3, 1, size=10_000),
                label=Flavours["bjets"].label,
                colour=Flavours["bjets"].colour,
                ratio_group="Ratio group 2",
            ),
            reference=True,
        )
        hist_plot.add(
            Histogram(
                rng.normal(3.5, 1, size=10_000),
                colour=Flavours["bjets"].colour,
                ratio_group="Ratio group 2",
                linestyle="--",
                linewidth=3,
                alpha=0.3,
            ),
        )
        hist_plot.draw()

        hist_plot.make_linestyle_legend(
            labels=["Reference", "Shifted"],
            linestyles=["solid", "dashed"],
            bbox_to_anchor=(0.25, 0.75),
            loc="upper right",
        )

        name = "test_ratio_groups.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")

        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1.4,
            )
        )

    def test_flavoured_labels(self):
        """Test different combinations of specifying the label when also specifying a
        flavour for the histogram.
        """
        rng = np.random.default_rng(seed=42)

        hist_plot = HistogramPlot(
            bins=100,
            atlas_brand=None,
            atlas_first_tag="",
            atlas_second_tag=(
                "Unit test plot to test the behaviour of \n"
                "the legend labels for different combinations \n"
                "of using the flavour label from the global \n"
                "config and not doing so"
            ),
            figsize=(8, 6),
        )
        # No flavour
        hist_plot.add(Histogram(rng.normal(0, 1, size=10_000), label="Unflavoured histogram"))
        # Flavour, but also label (using the default flavour label + the specified one)
        hist_plot.add(
            Histogram(
                rng.normal(4, 1, size=10_000),
                label="(flavoured, adding default flavour label '$b$-jets' to legend)",
                flavour="bjets",
            )
        )
        # Flavour + label (this time with suppressing the default flavour label)
        hist_plot.add(
            Histogram(
                rng.normal(8, 1, size=10_000),
                label="Flavoured histogram (default flavour label suppressed)",
                flavour="bjets",
                add_flavour_label=False,
                linestyle="--",
            )
        )
        # Flavoured, but using custom colour
        hist_plot.add(
            Histogram(
                rng.normal(12, 1, size=10_000),
                label="(flavoured, with custom colour)",
                flavour="bjets",
                linestyle="dotted",
                colour="b",
            )
        )
        hist_plot.draw()

        name = "test_flavoured_labels.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_weights(self):
        """Output plot with weights."""
        values = np.array([])
        values = np.array([0, 1, 2, 2, 3])
        weights = np.array([1, -1, 3, -2, 1])
        hist_exp = np.array([1, -1, 2])
        unc_exp = np.sqrt(np.array([1, (-1) ** 2, 3**2 + (-2) ** 2 + 1]))

        hist_plot = HistogramPlot(
            bins=3,
            figsize=(6, 5),
            atlas_first_tag=None,
            atlas_second_tag=(
                "Test plot for the behaviour of weights \n(both positive and negative)"
                f"\nExpected bin counts: {hist_exp}"
                f"\nExpected bin uncertainties: {unc_exp}"
            ),
            norm=False,
        )
        hist_plot.add(
            Histogram(
                values,
                weights=weights,
            )
        )
        hist_plot.draw()

        name = "test_histogram_weights.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_weights_wrong_shape(self):
        """Check if ValueError is raised if wieghts has."""
        values = np.array([0, 1, 2, 2, 3])
        weights = np.array([1, -1])

        with self.assertRaises(ValueError):
            Histogram(
                values,
                weights=weights,
            )

    def test_underoverflow_bin(self):
        """Test if under/overflow bins work as expected."""
        vals = [-1, 1, 2, 3, 6]
        vals_with_inf = [-1, 1, 2, 6, np.inf]

        hist_plot = HistogramPlot(bins=5, bins_range=(0, 5), underoverflow=False)
        hist_plot.atlas_second_tag = "This plot does not have under/overflow bins"
        hist_plot.add(Histogram(vals, colour="b"))
        hist_plot.add(Histogram(vals_with_inf, colour="r", linestyle="dotted"))
        hist_plot.draw()

        name = "test_histogram_without_underoverflow_bins.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

        hist_plot = HistogramPlot(bins=5, bins_range=(0, 5))
        hist_plot.atlas_second_tag = "This plot has under/overflow bins"
        hist_plot.add(Histogram(vals, colour="b"))
        hist_plot.add(Histogram(vals_with_inf, colour="r", linestyle="dotted"))
        hist_plot.draw()

        name = "test_histogram_with_underoverflow_bins.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")
        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")
        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_plot_filled_hist(self):
        bin_edges = [0, 1, 2, 3, 4, 5]
        bin_counts = [5, 4, 7, 12, 2]

        vals = [0, 1, 1, 5, 4, 2, 1, 3, 3, 5, 5, 5, 5, 5]

        hist_filled = Histogram(bin_counts, bin_edges=bin_edges)
        hist_notfilled = Histogram(vals)

        hist_plot = HistogramPlot(bins=bin_edges, underoverflow=False)
        hist_plot.add(hist_filled)
        hist_plot.add(hist_notfilled)

        hist_plot.draw()
        name = "test_filled_histogram.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")

        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")

        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )

    def test_plot_filled_hist_sumW2(self):
        bin_edges = [0, 1, 2, 3, 4, 5]
        bin_counts = [5, 4, 7, 12, 2]
        sum_squared_weights = [10, 7, 12, 21, 5]

        hist_plot = HistogramPlot(bins=bin_edges, underoverflow=False)
        hist_plot.add(
            Histogram(bin_counts, bin_edges=bin_edges, sum_squared_weights=sum_squared_weights)
        )

        hist_plot.draw()
        name = "test_filled_histogram_sumW2.png"
        hist_plot.savefig(f"{self.actual_plots_dir}/{name}")

        # Uncomment line below to update expected image
        # shutil.copy(f"{self.actual_plots_dir}/{name}", f"{self.expected_plots_dir}/{name}")

        self.assertIsNone(
            compare_images(
                f"{self.actual_plots_dir}/{name}",
                f"{self.expected_plots_dir}/{name}",
                tol=1,
            )
        )



"""Unit test script for the functions in utils/histogram.py."""

from __future__ import annotations

import unittest

import numpy as np
from testfixtures import LogCapture

from puma.utils import logger, set_log_level
from puma.utils.histogram import hist_ratio, hist_w_unc, save_divide

set_log_level(logger, "DEBUG")


class HistWUncTestCase(unittest.TestCase):
    """Test case for hist_w_unc function."""

    def setUp(self):
        self.input = np.array([0, 1, 1, 3])
        self.weights = np.array([1, 2, 1, 1])
        self.n_bins = 3
        self.bin_edges = np.array([0, 1, 2, 3])

        self.hist = np.array([1, 2, 1])
        self.unc = np.sqrt(np.array([1, 2, 1]))
        self.band = self.hist - self.unc

        self.hist_normed = np.array([1, 2, 1]) / len(self.input)
        self.unc_normed = np.sqrt(np.array([1, 2, 1])) / len(self.input)
        self.band_normed = self.hist_normed - self.unc_normed

        # --- weighted cases ---
        # 3 counts in second bin due to weights
        self.hist_weighted_normed = np.array([1, 3, 1]) / np.sum(self.weights)
        # use sqrt(sum of squared weights) for error calculation
        self.unc_weighted_normed = np.sqrt(np.array([1, 2**2 + 1, 1])) / np.sum(self.weights)
        self.band_weighted_normed = self.hist_weighted_normed - self.unc_weighted_normed

        self.hist_weighted = np.array([1, 3, 1])
        # use sqrt(sum of squared weights) for error calculation
        self.unc_weighted = np.sqrt(np.array([1, 2**2 + 1, 1]))
        self.band_weighted = self.hist_weighted - self.unc_weighted

    def test_under_overflow_values(self):
        """Test behaviour for under- and overflow values."""
        values_with_inf = np.array([-1, 1, 2, 100, np.inf])

        with self.subTest("Under/overflow values without under/overflow bins."):
            bins, hist, _, _ = hist_w_unc(values_with_inf, bins=5, bins_range=(0, 5))
            np.testing.assert_almost_equal(bins, np.linspace(0, 5, 6))
            # in this case only 40% of the values are shown in the plot
            np.testing.assert_almost_equal(hist, np.array([0, 0.2, 0.2, 0, 0]))

        with self.subTest("Under/overflow values with under/overflow bins."):
            bins, hist, _, _ = hist_w_unc(
                values_with_inf, bins=5, bins_range=(0, 5), underoverflow=True
            )
            np.testing.assert_almost_equal(bins, np.linspace(0, 5, 6))
            np.testing.assert_almost_equal(hist, np.array([0.2, 0.2, 0.2, 0, 0.4]))

    def test_hist_w_unc_zero_case(self):
        """Test what happens if empty array is provided as input."""
        bins, hist, unc, band = hist_w_unc(
            arr=[],
            bins=[],
        )

        np.testing.assert_almost_equal(bins, [])
        np.testing.assert_almost_equal(hist, [])
        np.testing.assert_almost_equal(unc, [])
        np.testing.assert_almost_equal(band, [])

    def test_hist_w_unc_normed(self):
        """Test normalised case."""
        bins, hist, unc, band = hist_w_unc(
            arr=self.input,
            bins=self.bin_edges,
        )

        np.testing.assert_almost_equal(bins, self.bin_edges)
        np.testing.assert_almost_equal(hist, self.hist_normed)
        np.testing.assert_almost_equal(unc, self.unc_normed)
        np.testing.assert_almost_equal(band, self.band_normed)

    def test_hist_w_unc_not_normed(self):
        """Test not normalised case."""
        bins, hist, unc, band = hist_w_unc(
            arr=self.input,
            bins=self.bin_edges,
            normed=False,
        )

        np.testing.assert_almost_equal(bins, self.bin_edges)
        np.testing.assert_almost_equal(hist, self.hist)
        np.testing.assert_almost_equal(unc, self.unc)
        np.testing.assert_almost_equal(band, self.band)

    def test_histogram_weighted_normalised(self):
        """Test weighted histogram (normalised)."""
        bin_edges, hist, unc, band = hist_w_unc(
            self.input, weights=self.weights, bins=self.n_bins, normed=True
        )

        np.testing.assert_array_almost_equal(self.bin_edges, bin_edges)
        np.testing.assert_array_almost_equal(self.hist_weighted_normed, hist)
        np.testing.assert_array_almost_equal(self.unc_weighted_normed, unc)
        np.testing.assert_array_almost_equal(self.band_weighted_normed, band)

    def test_histogram_weighted_not_normalised(self):
        """Test weighted histogram (not normalised)."""
        bin_edges, hist, unc, band = hist_w_unc(
            self.input, weights=self.weights, bins=self.n_bins, normed=False
        )

        np.testing.assert_array_almost_equal(self.bin_edges, bin_edges)
        np.testing.assert_array_almost_equal(self.hist_weighted, hist)
        np.testing.assert_array_almost_equal(self.unc_weighted, unc)
        np.testing.assert_array_almost_equal(self.band_weighted, band)

    def test_range_argument_ignored(self):
        """Test if the hist_range argument is ignored when bin_edges are provided."""
        bins_range = (1, 2)

        bin_edges, hist, _, _ = hist_w_unc(
            self.input,
            bins=self.bin_edges,
            bins_range=bins_range,
            normed=False,
        )

        # check if we end up with the same bin edges anyway
        np.testing.assert_array_almost_equal(self.bin_edges, bin_edges)
        np.testing.assert_array_almost_equal(self.hist, hist)

    def test_range_argument(self):
        """Test if the hist_range argument is used when bins is an integer."""
        # we test with range from 0 to 2, with 3 bins -> [0, 0.66, 1.33, 2] exp. bins
        bins_range = (0, 2)
        bins_exp = np.array([0, 2 / 3, 1 + 1 / 3, 2])
        hist_exp = np.array([1, 2, 0])

        bin_edges, hist, _, _ = hist_w_unc(
            self.input,
            bins=self.n_bins,
            bins_range=bins_range,
            normed=False,
        )

        # check if we end up with the same bin edges anyway
        np.testing.assert_array_almost_equal(bins_exp, bin_edges)
        np.testing.assert_array_almost_equal(hist_exp, hist)

    def test_negative_weights(self):
        """Test if negative weights are properly handled."""
        values = np.array([0, 1, 2, 2, 3])
        weights = np.array([1, -1, 3, -2, 1])

        hist_exp = np.array([1, -1, 2])
        # uncertainties are the sqrt(sum of squared weights)
        unc_exp = np.sqrt(np.array([1, (-1) ** 2, 3**2 + (-2) ** 2 + 1]))

        _, hist, unc, _ = hist_w_unc(values, weights=weights, bins=3, normed=False)
        np.testing.assert_array_almost_equal(hist_exp, hist)
        np.testing.assert_array_almost_equal(unc_exp, unc)

    def test_inf_treatment(self):
        """Test if infinity values are treated as expected."""
        values_with_infs = np.array([1, 2, 3, -np.inf, +np.inf, +np.inf])

        with self.subTest("Test if the warning for number of inf values is raised in hist_w_unc"):  # noqa: SIM117
            with LogCapture("puma") as log:
                _ = hist_w_unc(values_with_infs, bins=np.linspace(0, 3, 3))
                log.check((
                    "puma",
                    "WARNING",
                    "Histogram values contain 3 +-inf values!",
                ))
        with self.subTest(
            "Test if error is raised if inf values are in input but no range is defined"
        ), self.assertRaises(ValueError):
            hist_w_unc(values_with_infs, bins=10)

    def test_nan_check(self):
        """Test if the warning with number of nan values is raised in hist_w_unc."""
        values_with_nans = np.array([1, 2, 3, np.nan, np.nan])

        with LogCapture("puma") as log:
            _ = hist_w_unc(values_with_nans, bins=4)
            log.check((
                "puma",
                "WARNING",
                "Histogram values contain 2 nan values!",
            ))


class SaveDivideTestCase(unittest.TestCase):
    """Test case for save_divide function."""

    def test_zero_case(self):
        """Test zero case."""
        steps = save_divide(np.zeros(2), np.zeros(2))
        np.testing.assert_equal(steps, np.ones(2))

    def test_ones_case(self):
        """Test where all inputs are 1s."""
        steps = save_divide(np.ones(2), np.ones(2))
        np.testing.assert_equal(steps, np.ones(2))

    def test_half_case(self):
        """Test scenario where result is 1/2."""
        steps = save_divide(np.ones(2), 2 * np.ones(2))
        np.testing.assert_equal(steps, 0.5 * np.ones(2))

    def test_denominator_float(self):
        """Test scenario where denominator is a float."""
        steps = save_divide(np.ones(2), 2)
        np.testing.assert_equal(steps, 0.5 * np.ones(2))

    def test_numerator_float(self):
        """Test scenario where numerator is a float."""
        steps = save_divide(1, np.ones(2) * 2)
        np.testing.assert_equal(steps, 0.5 * np.ones(2))


class HistRatioTestCase(unittest.TestCase):
    """Test case for hist_ratio function."""

    def setUp(self):
        self.numerator = np.array([5, 3, 2, 5, 6, 2])
        self.denominator = np.array([3, 6, 2, 7, 10, 12])
        self.numerator_unc = np.array([0.5, 1, 0.3, 0.2, 0.5, 0.3])
        self.denominator_unc = np.array([1, 0.3, 2, 1, 5, 3])
        self.step = np.array([1.6666667, 1.6666667, 0.5, 1, 0.7142857, 0.6, 0.1666667])
        self.step_unc = np.array([
            0.16666667,
            0.16666667,
            0.16666667,
            0.15,
            0.02857143,
            0.05,
            0.025,
        ])
        self.step_rsd = np.array([4, 4, -5.1961524, 0, -4.8989795, -8, -11.8321596])
        self.step_rsd_unc = np.array([
            0.625,
            0.625,
            0.5773503,
            0,
            0.2041241,
            0.375,
            0.0507093,
        ])

    def test_hist_ratio(self):
        """Test if ratio is correctly calculated."""
        step, step_unc = hist_ratio(
            numerator=self.numerator,
            denominator=self.denominator,
            numerator_unc=self.numerator_unc,
        )

        np.testing.assert_almost_equal(step, self.step)
        np.testing.assert_almost_equal(step_unc, self.step_unc)

    def test_hist_rsd(self):
        step_rsd, step_rsd_unc = hist_ratio(
            numerator=self.numerator,
            denominator=self.denominator,
            numerator_unc=self.numerator_unc,
            method="root_square_diff",
        )

        np.testing.assert_almost_equal(step_rsd, self.step_rsd)
        np.testing.assert_almost_equal(step_rsd_unc, self.step_rsd_unc)

    def test_hist_not_same_length_numerator_denominator(self):
        """Test case where denominator and numerator have not the same length."""
        with self.assertRaises(AssertionError):
            _, _ = hist_ratio(
                numerator=np.ones(2),
                denominator=np.ones(3),
                numerator_unc=np.ones(3),
            )

    def test_hist_not_same_length_numerator_and_unc(self):
        """Test case where numerator and uncertainty have not the same length."""
        with self.assertRaises(AssertionError):
            _, _ = hist_ratio(
                numerator=np.ones(3),
                denominator=np.ones(3),
                numerator_unc=np.ones(2),
            )
